
############### RACE equations ################


### This module contains the main functions for
### the RACE (Retrieval by ACcumulating Evidence) equations for semantic memory retrieval

### Contact information:

###    o Internet

###        Web   : http://www.ai.rug.nl/~leendert/RACE.htm
###        E-mail: leendert@ai.rug.nl

###    o Postal address

###        Groningen University
###        ArtificiaL Intelligence and Cognitive Engineering
###        Drs. Leendert van Maanen
###        POBox 407
###        9700 AK Groningen
###        The Netherlands

###    o Tel. +31 50 3637603
###      Fax  +31 50 3636687


### global parameters ###
actr <- list()
actr$frequency <- 200 # recalibration frequency (Hz)
actr$d.context <- .72 # context decay
actr$ratio <- .8 # retrieval ratio
actr$growth <- .265 # growth scale parameter
actr$noise <- 0 # noise sd on total activation (A=B+C) 
actr$saliency <- 4.25 # parameter that controls stimulus input
actr$goal <- 1 # goal saliency
actr$default.base <- 0 # default baselevel activation
actr$refractory <- NULL #refractory period
actr$f <- 1 # scaling baselevel decay
actr$cutoff <- NULL # below cutoff, no more spreading of activation
nr.trials=10

### init functions ###
init.model <- function(duration, mem=NULL, task='BASIC') {
  ## this function initiates the parameters for a run of the RACE model
  set.range(duration)
  number.chunks <- set.memory(duration,mem, task)
  ## Initiate all activation matrices
  def.activation(duration,number.chunks)
}

set.range <- function(duration) {
  ## nr of time steps is calculated
  range <<- 1:duration
}

set.memory <- function(length.range,mem, task) {
  ## defines memory structure
  get.memory(mem)
  chunks$goal <<- task
  chunks$stimuli <<- NULL
  
  ## Make sure association matrix has no recurrent connections and is normalized
  diag(chunks$associations) <<- 0
  #chunks$associations <<- chunks$associations/max(chunks$associations)

  dimnames(chunks$associations) <<- list(chunks$all,chunks$all)
  number.chunks <- length(chunks$all)
  chunks$attended <<- matrix(NA, number.chunks, length.range)
  dimnames(chunks$attended) <<- list(chunks$all,NULL)

  return(number.chunks)
}

get.memory <- function(filename) {
  ## loads association matrix and predefined chunks from a given datafile
  
  ## this construction (instead of default argument) is because filename=NULL is passed as argument
  if (is.null(filename)) filename <- "DefaultDM.Rdat"
  load(filename)
  chunks <<- list()
  chunks$associations <<- assoc.matrix
  chunks$concepts <<- concepts
  chunks$lemmas <<- lemmas
  chunks$dm <<- dm
  chunks$all <<- all
}

def.activation <- function(length.range,number.chunks) {
  ## define activation matrices
  ## Matrix of the form m: chunks, n: time steps. Each cell contains the
  ## activation of that chunk on that point in time.

  activation <<- list()
  ## $all is sum of baselevel and context activation. Baselevel activation is the default
  ## ACT-R activation, context is the new evidence-based sampling activation.
  activation$all <<- matrix(actr$default.base, number.chunks,length.range,
                            dimnames=list(chunks$all,NULL))
  def.evidence(length.range,number.chunks)
}

def.evidence <- function(length.range,number.chunks) {
  ## define evidence accumulator matrices
  evidence <<- list()
  evidence$pos <<- matrix(0, number.chunks,length.range,
                            dimnames=list(chunks$all,NULL))
}

reset.chunks <- function(length.range,number.chunks) {
  ## resets the chunks list after a trial
  chunks$retrieved <<- NULL
  chunks$retrieved.lemmas <<- NULL

  chunks$stimuli <<- NULL
  
  chunks$attended <<- matrix(NA, number.chunks, length.range)
  dimnames(chunks$attended) <<- list(chunks$all,NULL)
}

### model functions ###

run.model <- function(start=1,end=duration, target.chunk, task) {
  ## Returns a tuple of form (time step,chunk)
  number.chunks <- length(chunks$all)
  for (timestep in start:end) {
    retrieval <- update.activation(timestep, number.chunks, target.chunk, task)
    if (!(is.null(retrieval))) break
  }
  return(list(RT=timestep,chunk=retrieval))
}

update.activation <- function(timestep,number.chunks, target.chunk, task) {
  ## for every time step, this function recalibrates the activation and updates
  ## the sensory buffer contents

  ## update the sensory buffers
  check.stimuli(timestep)
  #check.retrievedlemmas(timestep)

  ## if a stimulus is present...
  if (!(is.null(chunks$stimuli))) {
    ## calculate associative matrix
    assoc <- set.associative.strength(timestep)
    assoc <- stimulus.act(assoc)
    #assoc <- goal.act(assoc,task)

    ## calculate (positive) spreading activation
    set.evidence(assoc,timestep)
  }

  ## calculate net evidence
  evidence <- check.evidence(timestep)

  ## calculate context and baselevel activation for current time step
  activation$all[,timestep] <<- rnorm(number.chunks,
                                      context.activation(timestep,evidence),
                                      actr$noise)
  
  ## check and handle retrieval of chunks
  retrieval <- process.retrieval(timestep, target.chunk)
  return(retrieval)
}

context.activation <- function(timestep, evidence) {
  ## context activation is the volatile activation that regulates retrieval
  if (timestep>1) {
    act <- actr$d.context*activation$all[,timestep-1] + actr$growth*evidence
    act[abs(act)==Inf] <- 0
  } else act <- 0

  ## context activation can only increase total activation
  return(pmax(0,act))
}

check.stimuli <- function(timestep) {
  ## checks if a stimulus should be placed in the sensory buffers on this time step
  on <- chunks$stimulus.onset==timestep
  if (any(on==TRUE)) {
    chunks$stimuli <<- c(chunks$stimuli, names(chunks$stimulus.onset[on]))
    #chunks$retrieved <<- NULL
  }
  
  off <- chunks$stimulus.off==timestep
  if (any(off==TRUE)) {
    ## turn stimuli off after the duration
    chunks$stimuli <<- chunks$stimuli[which(off==FALSE)]
    if (any(is.na(chunks$stimuli))) chunks$stimuli <<- NULL
  }
}

check.retrievedlemmas <- function(timestep) {
  ## for which chunks the refreactory period is over?
  compete <- chunks$all[!(is.na(chunks$attended[,timestep-actr$refractory]))]
  chunks$retrieved.lemmas <<- chunks$retrieved.lemmas[!(chunks$retrieved.lemmas %in% compete)]                                
}
    
set.associative.strength <- function(timestep) {
  ## Spreading activation mechanism.
  ## initial activation values are taken from default baselevel parameter,
  ## afterwards the activation at t-1 is taken
  if (timestep==1) {
    default.activation <- rep(actr$default.base,length(chunks$all))
    assoc <- default.activation * chunks$associations
  } else {
    assoc <- activation$all[,timestep-1] * chunks$associations

    ## NA's are replaced by 0 for computational reasons
    assoc[is.na(assoc)] <- 0
  }
  return(assoc)
}

stimulus.act <- function(assoc) {
  ## spreading activation to chunks associated with present stimuli is increased

  ## exclude lemmas that are already retrieved
  if (!(is.null(chunks$retrieved))) {
    #browser()
    competing <- chunks$associations[chunks$retrieved,]==0
    ##competing <- chunks$all[!(chunks$all %in% chunks$retrieved)]
  } else competing <- chunks$dm
  
  assoc[chunks$stimuli,competing] <- assoc[chunks$stimuli,competing] +
    chunks$associations[chunks$stimuli,competing] * actr$saliency
  
  assoc[competing,chunks$stimuli] <- assoc[competing,chunks$stimuli] +
    chunks$associations[competing,chunks$stimuli] * actr$saliency
  
  return(assoc)
}

goal.act <- function(assoc, task) {
  ## spreading activation to chunks associated with present stimuli is increased

  ## exclude lemmas that are already retrieved
  #competing <- names(which(chunks$associations[chunks$retrieved,]==0))
  #competing <- chunks$all[!(chunks$all %in% chunks$retrieved)]
  competing <- chunks$dm
  
  assoc[task,competing] <- assoc[task,competing] +
    chunks$associations[task,competing] * actr$goal
  
  assoc[competing,task] <- assoc[competing,task] +
    chunks$associations[competing,task] * actr$goal
  
  return(assoc)
}

set.evidence <- function(assoc, timestep) {
  #competing <- chunks$dm[!(chunks$dm %in% chunks$retrieved)]
  #comp.assoc <- assoc[,competing]
  
  if (is.matrix(assoc)) {
    evidence$pos[,timestep] <<- colSums(assoc, na.rm=TRUE)
  } else {
    evidence$pos[,timestep] <<- sum(assoc, na.rm=TRUE)
  }
}

check.evidence <- function(timestep) {
  ## calculates for a certain timestep the net evidence per chunk
  net.evidence <- evidence$pos[,timestep]#-evidence$neg[,timestep]
  return(net.evidence)
}

process.retrieval <- function(timestep, target.chunk) {
  ## check for retrieval and, if so, add a presentation of the retrieved chunk
  ## only chunks that match the slots in the retrieval request - in this case,
  ## chunks of the lemma chunk type - will be retrieved.
  retrieved.lemmas <- check.retrieval(timestep,"lemmas")
  retrieved.concepts <- NULL#check.retrieval(timestep,"concepts")
  retrieved.mappings <- check.retrieval(timestep,"mappings")
  retrieved.chunks <- c(retrieved.concepts,retrieved.lemmas, retrieved.mappings)
  #if (length(retrieved.lemmas)>0 &&
  #    all(!(is.na(retrieved.lemmas)))) {
  #  chunks$retrieved.lemmas <<- c(chunks$retrieved.lemmas,retrieved.lemmas)
  #}
    
  if (length(retrieved.chunks)>0 &&
      all(!(is.na(retrieved.chunks)))) {
    chunks$retrieved <<- unique(c(chunks$retrieved,retrieved.chunks))
    add.presentation(timestep, retrieved.chunks)
    
    return(retrieved.chunks)
  } else return(NULL)
}

check.retrieval <- function(timestep, chunktype) {
  ## check if a chunk in dm is retrieved, ie has crossed the context retrieval threshold
  #competing.chunks <- chunks[[chunktype]][!(chunks[[chunktype]] %in% chunks$retrieved)]
  competing.chunks <- chunks[[chunktype]]
  
  competing.act <- activation$all[competing.chunks,timestep]

  ratio <- luce.ratio(competing.act)
  retrieved.chunks <- competing.chunks[which(ratio>=actr$ratio)]
  
  return(retrieved.chunks)
}

luce.ratio <- function(act) {
  ## Luce's ratio: exponentially scaled activations give the probability that a chunk is retrieved.
  exp(act)/sum(exp(act))
}

reset.retrieval <- function(chunks,timestep) {
  ## reset the context activation to default value (0) if achunk is retrieved
  activation$context[chunks,timestep] <<- 0
}

add.presentation <- function(timestep,chunk) {
  ## add in the attended matrix a marker that a chunk has been attended, for use
  ## in baselevel activation equation
  chunks$attended[chunk,timestep] <<- timestep
}

### experiment functions ###

latency <- function(act, ret.onset=cur.ret.onset) {
  ## latency equation for RACE
  above <- which(act>=actr$ratio)
  above <- above[above>ret.onset][1]
  above <- (above-ret.onset)/actr$frequency
  return(above)
}


do.trials <- function(stimuli.list="house",  # stimuli.list takes either one or two chunks of the form c(distactor,target)
                      target.chunk="lhouse", # chunk to be retrieved
                      SOA.list=0,            # difference in onset time between cue en distractor. Can be a list for different conditions
                      stimulus.durations=NULL,  # duration of stimuli. NULL means continouous stimulus as of stimulus onset. Can be either on or two durations (depending on same or different duration for cue and distractor) 
                      nr.trials=1,
                      maxRT=2,               # max time a trial will run until a retrieval failure is recorded
                      task='BASIC',          # BASIC level naming or CATEGORY naming
                      mem='DefaultDM.Rdat'   # file contains chunks involved in experiment and association matrix
                      ) {
  ## do one trial of memory retrieval

  RT <- matrix(NA,length(SOA.list),nr.trials,dimnames=list(SOA.list,NULL))
  SOA.list <- ceiling(SOA.list*actr$frequency)
  
  for (SOA in SOA.list) {
    duration <- abs(SOA)+maxRT*actr$frequency # default 2s (maxRT) after target stimulus onset a trial is ended
    init.model(duration, mem, task)
    min.onset <- 1
    number.chunks <- length(chunks$all)

    ## shift time axis and calculate stimuli onset
    if (length(stimuli.list)==2) {
      if (SOA<0) chunks$stimulus.onset <<- c(min.onset,abs(SOA)+1)
      if (SOA>=0) chunks$stimulus.onset <<- c(SOA+1,min.onset)
    } else chunks$stimulus.onset <<- min.onset
    names(chunks$stimulus.onset) <<- stimuli.list
    chunks$stimulus.off <<- chunks$stimulus.onset + ceiling(stimulus.durations*actr$frequency)

    for (trial in 1:nr.trials) {
      ## calculate latency for this trial and SOA
      RT[which(SOA.list==SOA),trial] <- (do.trial(stimuli.list, target.chunk, duration, number.chunks, task) -
                 chunks$stimulus.onset[length(chunks$stimulus.onset)]) /
                   actr$frequency
    }
  }
return(RT)
}

do.trial <- function(stimuli.list, target.chunk, duration, number.chunks, task) {

  def.activation(duration, number.chunks)
  reset.chunks(duration,number.chunks)
  
  retrieve <- list(RT=0,chunk=NA) #tuple (timestep,chunk) returned by run.model; RT=0, because run.model takes RT+1 as input
  
  ## while the target chunk is not yet retrieved:
  while (!(target.chunk %in% retrieve$chunk) &&
         retrieve$RT<duration) {
    ## run RACE until next context threshold crossing
    retrieve <- run.model(retrieve$RT+1,duration, target.chunk, task)
    ## if the target is retrieved before stimulus onset, then clear retrieval buffer
    if (retrieve$RT<=chunks$stimulus.onset[length(chunks$stimulus.onset)]) {
      retrieve$chunk <- NA
    }
  }
  ## retrieval failure is encoded as NA 
  if (retrieve$RT>=duration-1) retrieve$RT=NA
  return(retrieve$RT)
}

def.mem <- function(filename="DefaultDMCAT.Rdat") {
  ## write default memory
  assoc.matrix <- chunks$associations
  concepts <- chunks$concepts
  lemmas <- chunks$lemmas
  dm <- chunks$dm
  all <- chunks$all
  stimuli <- chunks$all[!(chunks$all %in% chunks$dm)]
  save(file=filename,assoc.matrix,concepts,lemmas,dm,all,stimuli)
}

add.dm <- function(chunk, type, assocs, mem="DMnew.Rdat") {
  get.memory(mem)
  #browser()
  chunks$associations <<- matrix(c(chunks$associations,
                                   assocs[1:dim(chunks$associations)[1],1:length(chunk)]),
                                 dim(chunks$associations))
  chunks$associations <<- matrix(c(t(chunks$associations),
                                   assocs),
                                 dim(assocs))
  chunks[[type]] <<- c(chunks[[type]],chunk)
  chunks$dm <<- c(chunks$dm,chunk)
  chunks$all <<- c(chunks$all,chunk)
  dimnames(chunks$associations) <<- list(chunks$all,chunks$all)
}

recall.probability <- function(act) {
  ratio <- luce.ratio(act)
  p <- 1/(1+ratio/actr$ratio)
  return(p)
}
