############### Subliminal priming model using RACE equations ################


### This module contains the function to run the subliminal priming model
### using the race equations provided in race.r
### The model results are published in Van Maanen, L., Van Rijn, H. (2007)
### Accounting for subliminal priming in ACT-R., in proceedings of the 8th
### International Conference on Cognitive Modeling, Ann Arbor, MI, USA.

### Contact information:

###    o Internet

###        Web   : http://www.ai.rug.nl/~leendert/RACE.htm
###        E-mail: leendert@ai.rug.nl

###    o Postal address

###        Groningen University
###        ArtificiaL Intelligence and Cognitive Engineering
###        Drs. Leendert van Maanen
###        POBox 407
###        9700 AK Groningen
###        The Netherlands

###    o Tel. +31 50 3637603
###      Fax  +31 50 3636687
if (F) {

  ###run this code to initiate the subliminal priming model and run some trials
  
  chunks$associations <- matrix(0,12,12)
  chunks$lemmas <- c("lred","lgreen","lunrel")
  chunks$concepts <- c("RED","GREEN","UNREL")
  chunks$mappings <- c("mmred","mmgreen")
  chunks$dm <- c(chunks$concepts,chunks$lemmas,chunks$mappings)
  chunks$all <- c(chunks$dm,"cred","wred","wgreen","wunrel")
  dimnames(chunks$associations) <- list(chunks$all,chunks$all)

  ## concepts - concepts
  chunks$associations["RED","GREEN"] <- .2
  chunks$associations["GREEN","RED"] <- .2

  ## concepts - lemmas
  chunks$associations["RED","lred"] <- .3
  chunks$associations["lred","RED"] <- .3
  chunks$associations["lgreen","GREEN"] <- .3
  chunks$associations["GREEN","lgreen"] <- .3
  chunks$associations["lunrel","UNREL"] <- .3
  chunks$associations["UNREL","lunrel"] <- .3

  ## concepts - motor mappings
  chunks$associations["RED","mmred"] <- .5
  chunks$associations["mmred","RED"] <- .5
  chunks$associations["mmgreen","GREEN"] <- .5
  chunks$associations["GREEN","mmgreen"] <- .5

  ## concepts - colors
  chunks$associations["RED","cred"] <- .4
  chunks$associations["cred","RED"] <- .4

  ## lemmas - words
  chunks$associations["wred","lred"] <- .63
  chunks$associations["lred","wred"] <- .63
  chunks$associations["lgreen","wgreen"] <- .63
  chunks$associations["wgreen","lgreen"] <- .63
  chunks$associations["lunrel","wunrel"] <- .63
  chunks$associations["wunrel","lunrel"] <- .63

  def.mem('sublim6.Rdat')

  summaryandplot(stimuli=c("wgreen","cred"),
                 target="mmred",
                 SOA=0,
                 stimulus.durations=c(.075,2),
                 maxRT=.1,
                 mem='sublim5.Rdat')

  prediction["unaware",] <- 1000*sublim(stimulus.durations=c(.075,2), SOA.list=c(0),mem='sublim5.Rdat')# unaware of prime
  prediction["aware",] <- 1000*sublim(stimulus.durations=c(.4,2), SOA.list=c(0),mem='sublim5.Rdat') # aware


  r2 <- cor(c(sublim.data-sublim.data[,'noword']), c(prediction-prediction[,'noword']))^2
}

### global parameters ###
actr <- list()
actr$frequency <- 200 # recalibration frequency (Hz)
actr$d.context <- .72 # context decay
actr$ratio <- .81 # retrieval ratio
actr$growth <- .255 # growth scale parameter
actr$noise <- 0 # noise sd on total activation (A=B+C) 
actr$saliency <- 3 # parameter that controls stimulus input
actr$goal <- 1 # goal saliency
actr$default.base <- 0 # default baselevel activation
actr$refractory <- NULL #refractory period
actr$f <- 1 # scaling baselevel decay
actr$cutoff <- NULL # below cutoff, no more spreading of activation
nr.trials=10

process.retrieval <- function(timestep, target.chunk) {
  ## check for retrieval and, if so, add a presentation of the retrieved chunk
  ## only chunks that match the slots in the retrieval request - in this case,
  ## chunks of the lemma chunk type - will be retrieved.
  retrieved.lemmas <- check.retrieval(timestep,"lemmas")
  retrieved.concepts <- check.retrieval(timestep,"concepts")
  retrieved.mappings <- check.retrieval(timestep,"mappings")
  retrieved.chunks <- c(retrieved.concepts,retrieved.lemmas, retrieved.mappings)
  
  if (length(retrieved.mappings)>0 &&
      all(!(is.na(retrieved.chunks)))) {
    chunks$retrieved <<- c(chunks$retrieved,retrieved.mappings)
    add.presentation(timestep, retrieved.mappings)
    
    return(retrieved.chunks)
  } else return(NULL)
}

def.mem <- function(filename="DefaultDMCAT.Rdat") {
  ## write default memory
  assoc.matrix <- chunks$associations
  concepts <- chunks$concepts
  lemmas <- chunks$lemmas
  mappings <- chunks$mappings
  dm <- chunks$dm
  all <- chunks$all
  stimuli <- chunks$all[!(chunks$all %in% chunks$dm)]
  save(file=filename,assoc.matrix,concepts,lemmas,mappings,dm,all,stimuli)
}

get.memory <- function(filename) {
  ## loads association matrix and predefined chunks from a given datafile
  
  ## this construction (instead of default argument) is because filename=NULL is passed as argument
  if (is.null(filename)) filename <- "DefaultDM.Rdat"
  load(filename)
  chunks <<- list()
  chunks$associations <<- assoc.matrix
  chunks$concepts <<- concepts
  chunks$lemmas <<- lemmas
  chunks$mappings <<- mappings
  chunks$dm <<- dm
  chunks$all <<- all
}

sublim <- function(stimulus.durations=NULL,SOA.list=0,nr.trials=1,mem='sublim.Rdat') {
  
  conditionset <- list(noword=c("cred","cred"),
                       neutral=c("wunrel","cred"),
                       congruent=c("wred","cred"),
                       incongruent=c("wgreen","cred"))
  sublim.data <- array(NA, dim=c(length(conditionset),nr.trials,length(SOA.list)),
                       dimnames=list(names(conditionset),NULL,SOA.list))
  
  for (SOA in SOA.list) {
    for (c in 1:length(conditionset)) {
      condition <- conditionset[c]
      sublim.data[names(condition),,which(SOA.list==SOA)] <-
        do.trials(stimuli.list=unlist(condition),target.chunk="mmred",SOA,
                  stimulus.durations=stimulus.durations,nr.trials=nr.trials, mem=mem)
      cat(names(condition),"lemmas",check.aware(chunktype="lemmas"),"\n")
    }	
  }    	
  return(sublim.data)
}

check.aware <- function(chunktype="lemma", act=activation$all, range=NA, chunk=NA) {
  ## This function checks if a chunk is within awareness, that is, is retrieved (and
  ## placed in a buffer). By default, it checks whether any chunk of the lemma type
  ## is retrieved during a trial.

  ## chunktype: chunktype to be checked
  ## act: activation over time of the chunks to be checked
  ## range: time interval for which check should be, NA means complete range
  ## chunk: specific chunk you want to check. Should be of the specified chunktype
  
  if (any(is.na(range))) {
    comp.act <- act[chunks[[chunktype]],]
  } else comp.act <- act[chunks[[chunktype]],range]

  ratios <- apply(comp.act,2,luce.ratio)

  if (is.na(chunk)) {
    return(any(ratios>=actr$ratio))
  } else return(any(ratios[chunk,]>=actr$ratio))
}

limen <- function(mem="sublim4.Rdat") {
  ## this function determines the awareness threshold. It checks the time
  ## needed for a single chunk to be retrieved from memory 
  do.trials(stimuli.list="wred", target.chunk="lred", mem=mem, maxRT=5)
}
