
############### Picture-Word Interference model using RACE equations ################


### This module contains the main functions to calculate the expected latencies in a
### picture-word interference experiment. The model is meant as a demonstration of
### the RACE (Retrieval by ACcumulating Evidence) equations for semantic memory retrieval
### and is part of the following publication:

### Van Maanen, L. & Van Rijn (2006), H., An Accumulator Model Account of Semantic
### Interference in Memory Retrieval, submitted.

### Contact information:

###    o Internet

###        Web   : http://www.ai.rug.nl/~leendert/
###        E-mail: leendert@ai.rug.nl

###    o Postal address

###        University of Groningen
###        Artificial Intelligence Department
###        Drs. Leendert van Maanen
###        Grote Kruisstraat 2/1
###        9712 TS Groningen
###        The Netherlands

###    o Tel. +31 50 3637603
###      Fax  +31 50 3636687


### global parameters ###
actr <- list()
actr$d.base <- .5 # ACT-R baselevel decay
actr$frequency <- 200 # recalibration frequency (Hz)
actr$d.context <- .1 # context decay
actr$t.context <- 2 # context threshold
actr$growth <- .1 # growth scale parameter
actr$noise <- 0 # noise sd on total activation (A=B+C) 
actr$saliency <- 15 # parameter that controls stimulus input.
actr$default.base <- -3 # default baselevel activation

### init functions ###
init.model <- function(duration, mem=NULL) {
  ## this function initiates the parameters for a run of the RACE model
  ## duration in seconds
  length.range <- set.range(duration)
  number.chunks <- set.memory(length.range,mem)
  ## Initiate all activation matrices
  def.activation(length.range,number.chunks)
}

set.range <- function(duration) {
  ## nr of time steps is calculated
  range <<- 1:duration
  return(length(range))
}

set.memory <- function(length.range,mem) {
  ## defines memory structure
  get.memory(mem)
  chunks$stimuli <<- NULL
  
  ## Make sure association matrix has no recurrent connections and is normalized
  diag(chunks$associations) <<- 0
  chunks$associations <<- chunks$associations/max(chunks$associations)

  dimnames(chunks$associations) <<- list(chunks$all,chunks$all)
  number.chunks <- length(chunks$all)
  chunks$attended <<- matrix(NA, number.chunks, length.range)
  dimnames(chunks$attended) <<- list(chunks$all,NULL)
  
  return(number.chunks)
}

get.memory <- function(filename) {
  ## loads association matrix and predefined chunks from a given datafile
  
  ## this construction (instead of default argument) is because filename=NULL is passed as argument
  if (is.null(filename)) filename <- "DefaultDM.Rdat"
  load(filename)
  chunks <<- list()
  chunks$associations <<- defaultAssocMatrix
  chunks$concepts <<- concepts
  chunks$lemmas <<- lemmas
  chunks$icons <<- icons
  chunks$dm <<- c(chunks$concepts,chunks$lemmas,chunks$icons)
  chunks$all <<- c(chunks$dm,stimuli)
}  

def.activation <- function(length.range,number.chunks) {
  ## define activation matrices
  ## Matrix of the form m: chunks, n: time steps. Each cell contains the
  ## activation of that chunk on that point in time.

  activation <<- list()
  ## $all is sum of baselevel and context activation. Baselevel activation is the default
  ## ACT-R activation, context is the new evidence-based sampling activation.
  activation$all <<- matrix(actr$default.base, number.chunks,length.range,
                            dimnames=list(chunks$all,NULL))
  activation$base <<- activation$all
  activation$context <<- matrix(0, number.chunks,length.range,
                             dimnames=list(chunks$all,NULL))
  def.evidence()
}

def.evidence <- function() {
  ## define evidence accumulator matrices
  evidence <<- list()
  
  ## Copy structure of activation$context:
  evidence$pos <<- activation$context
  evidence$neg <<- activation$context
}

reset.chunks <- function(length.range,number.chunks) {
  ## resets the chunks list after a trial
  chunks$retrieved <<- NULL
  chunks$stimuli <<- NULL
  
  chunks$attended <<- matrix(NA, number.chunks, length.range)
  dimnames(chunks$attended) <<- list(chunks$all,NULL)
}

### model functions ###

run.model <- function(start=1,end=duration, target.chunk) {
  ## Returns a tuple of form (time step,chunk)
  number.chunks <- length(chunks$all)
  for (timestep in start:end) {
    retrieval <- update.activation(timestep, number.chunks, target.chunk)
    if (!(is.null(retrieval))) break
  }
  return(list(RT=timestep,chunk=retrieval))
}

update.activation <- function(timestep,number.chunks, target.chunk) {
  ## for every time step, this function recalibrates the activation and updates
  ## the sensory buffer contents

  ## update the sensory buffers
  check.stimuli(timestep, target.chunk)

  ## if a stimulus is present...
  if (!(is.null(chunks$stimuli))) {
    ## calculate associative matrix
    assoc <- set.associative.strength(timestep)

    ## calculate positive and negative spreading activation
    set.positive.evidence(assoc,timestep)
    set.negative.evidence(assoc,timestep)
  }

  ## calculate net evidence
  evidence <- check.evidence(timestep)

  ## calculate context and baselevel activation for current time step
  activation$context[,timestep] <<- context.activation(timestep,evidence)
  activation$base[,timestep] <<- baselevel.activation(timestep)

  ## calculate activation and add gaussian noise. Activation
  ## should not be below default baselevel
  activation$all[,timestep] <<- pmax(actr$default.base,
                                     rnorm(number.chunks,activation$context[,timestep] +
                                           activation$base[,timestep], actr$noise))
  
  ## check and handle retrieval of chunks
  retrieval <- process.retrieval(timestep)
  return(retrieval)
}

baselevel.activation <- function(time, decay=actr$d.base) {
  ## standard ACT-R baselevel activation function (Anderson & Lebiere, 1998)
  ## this equation applies to seconds, so the timescale has to be recalculated:
  attended <- chunks$attended/actr$frequency
  time <- time/actr$frequency
  if (is.matrix(attended)) {
    activation <- log(rowSums((time-attended)^(-1*decay),na.rm=TRUE))
  } else {
    activation <- log(sum((time-attended)^(-1*decay),na.rm=TRUE))
      }
  activation[activation==-Inf] <- actr$default.base
  return(activation)
}

context.activation <- function(timestep, evidence, accumulation.start=evidence$accumulation.start,
                               decay=actr$d.context) {

  ## context activation is the volatile activation that regulates retrieval
  if (timestep>1) {
    age <- timestep
    act <- activation$context[,timestep-1] +
      actr$growth*(exp(evidence)-1) - 
      decay*log(age)/actr$frequency
    
    act[abs(act)==Inf] <- 0
  } else act <- 0

  ## context activation can only increase total activation
  return(pmax(0,act))
}

check.stimuli <- function(timestep, target.chunk) { 
  ## checks if a stimulus should be placed in the sensory buffers on this time step
  on <- chunks$stimulus.onset==timestep
  
  if (any(on==TRUE)) {
    chunks$stimuli <<- c(chunks$stimuli, names(chunks$stimulus.onset[on]))
    
    ## IF the target chunk is retrieved (before target stimulus onset)
    ## THEN remove it from the retrieved chunks list (so it can be retrievedd again)
    if (any(chunks$retrieved==target.chunk)) {
      chunks$retrieved <<- chunks$retrieved[chunks$retrieved!=target.chunk]
    }
  }
}

    
set.associative.strength <- function(timestep) {
  ## the exp of the associative strength is taken to scale the strenghts from 0 to Inf
  ## initial activation values are taken from default baselevel parameter,
  ## afterwards the activation at t-1 is taken
  if (timestep==1) {
    default.activation <- rep(actr$default.base,length(chunks$all))
    assoc <- exp(default.activation) * chunks$associations
  } else {
    assoc <- exp(activation$all[,timestep-1]) * chunks$associations
    assoc[is.na(assoc)] <- 0
  }

  ## spreading activation to chunks associated with present stimuli is increased
  assoc[chunks$stimuli,] <- assoc[chunks$stimuli,]*actr$saliency 
  assoc[,chunks$stimuli] < assoc[,chunks$stimuli]*actr$saliency
  
  return(assoc)
}

set.positive.evidence <- function(assoc,timestep) {
  ## chunks of different chunk types spread positive activation
  
  ## define which concept chunks spread activation (already retrieved not)
  concepts <- which(chunks$all %in% chunks$concepts & !(chunks$all %in% chunks$retrieved))

  ## define which other chunks spread activation
  lemmas.icons <- which(chunks$all %in% c(chunks$lemmas,chunks$icons) & !(chunks$all %in% chunks$retrieved))

  ## calc associative matrices
  assoc.lemmasicons <- assoc[!(chunks$all %in% c(chunks$lemmas,chunks$icons)),lemmas.icons]
  assoc.concepts <- assoc[c(chunks$lemmas,chunks$icons),concepts]

  ## calc total association per chunk. different computations if nr. chunks is 1 or number of
  ## sources of activation is 1
  if (is.matrix(assoc.lemmasicons)) {
    evidence$pos[lemmas.icons,timestep] <<- colSums(assoc.lemmasicons, na.rm=TRUE)
  } else if (length(lemmas.icons)>1) {
    evidence$pos[chunks$dm,timestep] <<- assoc.lemmasicons
  } else {
    evidence$pos[chunks$dm,timestep] <<- sum(assoc.lemmasicons)
  }
  if (is.matrix(assoc.concepts)) {
    evidence$pos[concepts,timestep] <<- colSums(assoc.concepts, na.rm=TRUE)
  } else if (length(concepts)>1) {
    evidence$pos[concepts,timestep] <<- assoc.concepts
  } else {
    evidence$pos[concepts,timestep] <<- sum(assoc.concepts)
  }
}

set.negative.evidence <- function(assoc,timestep) {
  ## chunk of the same type spread negative activation
  ## Since there are no recurrent connections, we can assume that all chunks
  ## in DM that have associations spread negative activation.
  concept.assoc <- assoc[chunks$concepts,chunks$concepts]
  lemma.assoc <- assoc[chunks$lemmas,chunks$lemmas]
  icon.assoc <- assoc[chunks$icons,chunks$icons]
  if (is.matrix(concept.assoc)) {
    evidence$neg[chunks$concepts,timestep] <<- colSums(concept.assoc,na.rm=TRUE)
  } else {
    ## Only occurs when we have just one concept in DM... (hmmmm.... :-))
    evidence$neg[chunks$concepts,timestep] <<- 0
  }
  if (is.matrix(lemma.assoc)) {
    evidence$neg[chunks$lemmas,timestep] <<- colSums(lemma.assoc,na.rm=TRUE)
  } else {
    evidence$neg[chunks$lemmas,timestep] <<- 0
  }
  if (is.matrix(icon.assoc)) {
    evidence$neg[chunks$icons,timestep] <<- colSums(icon.assoc,na.rm=TRUE)
  } else {
    evidence$neg[chunks$icons,timestep] <<- 0
  }
}

check.evidence <- function(timestep) {
  ## calculates for a certain timestep the net evidence per chunk
  net.evidence <- evidence$pos[,timestep]-evidence$neg[,timestep]
  return(net.evidence)
}

process.retrieval <- function(timestep) {
  ## check for retrieval and, if so, add a presentation of the retrieved chunk
  retrieved.chunks <- check.retrieval(timestep)

  if (length(retrieved.chunks)>0 &&
      all(!(is.na(retrieved.chunks)))) {
    chunks$retrieved <<- c(chunks$retrieved,retrieved.chunks)
    add.presentation(timestep, retrieved.chunks)

    ## the retrieval process is reset
    reset.retrieval(retrieved.chunks,timestep)
    
    return(retrieved.chunks)
  } else return(NULL)
}

check.retrieval <- function(timestep) {
  ## check if a chunk in dm is retrieved, ie has crossed the context retrieval threshold
  competing.chunks <- chunks$dm[!(chunks$dm %in% chunks$retrieved)]
  retrieved.chunks <- competing.chunks[activation$all[competing.chunks,timestep]>=actr$t.context]
  return(retrieved.chunks)
}

reset.retrieval <- function(chunks,timestep) {
  ## reset the context activation to default value (0) if achunk is retrieved
  activation$context[chunks,timestep] <<- 0
}

add.presentation <- function(timestep,chunk) {
  ## add in the attended matrix a marker that a chunk has been attended, for use
  ## in baselevel activation equation
  chunks$attended[chunk,timestep] <<- timestep
}

### experiment functions ###

latency <- function(act, ret.onset=cur.ret.onset) {
  ## latency equation for RACE
  above <- which(act>=actr$t.context)
  above <- above[above>ret.onset][1]
  above <- (above-ret.onset)/actr$frequency
  return(above)
}

## Glaser & Dungelhoff (1984):
## the function diff.SOAs() replicates the Picture Word Interference data
## from Experiment 1 Glaser & Dungelhoff (1984).

diff.SOAs <- function(SOA.list=c(-.4,-.3,-.2,-.1,0,.1,.2,.3,.4),
                      nr.trials=1,
                      plotting=TRUE,
                      filename=NULL) {
  
  ## record date and time for tracking purposes
  date.and.time <- date()

  ## define data structure
  picture.naming <<- list()
  picture.naming$concept.congruent <<- matrix(NA,length(SOA.list),nr.trials)
  picture.naming$category.congruent <<- matrix(NA,length(SOA.list),nr.trials)
  picture.naming$incongruent <<- matrix(NA,length(SOA.list),nr.trials)
  picture.naming$neutral <<- rep(NA,nr.trials)

  word.reading <<- list()
  word.reading$concept.congruent <<- matrix(NA,length(SOA.list),nr.trials)
  word.reading$category.congruent <<- matrix(NA,length(SOA.list),nr.trials)
  word.reading$incongruent <<- matrix(NA,length(SOA.list),nr.trials)
  word.reading$neutral <<- rep(NA,nr.trials)

  if (plotting) {
    par(mfrow=c(1,2))
    plot(SOA.list,SOA.list,type="n",ylim=c(-.1,.2), main="Picture Naming",
         ylab="latency difference", xlab="SOA")
    legend(0,.2,pch=c(15, 16, 17), legend=c("Incongruent", "Category Congruent", "Concept Congruent"))
  }

  
  ## picture naming task, neutral condition
  picture.naming$neutral <<- do.trial(stimuli.list="phouse",target.chunk="lhouse",
                                                SOA.list,nr.trials)
  if (plotting) {
    abline(h=0, lty=2)
    neutral <-  mean(picture.naming$neutral, na.rm=TRUE)
  }

  ## concept congruent condition
  picture.naming$concept.congruent <<- do.trial(
                                                       stimuli.list=c("house","phouse"),
                                                        target.chunk="lhouse",SOA.list,nr.trials)
  if (plotting) lines(SOA.list,rowMeans(picture.naming$concept.congruent, na.rm=TRUE) - neutral,
                      type="b", pch=17)

  ## category congruent condition
  picture.naming$category.congruent <<- do.trial(stimuli.list=c("castle","phouse"),
                                                         target.chunk="lhouse",SOA.list,nr.trials)
  if (plotting) lines(SOA.list,rowMeans(picture.naming$category.congruent, na.rm=TRUE) - neutral,
                      type="b", pch=16)

  ## incongruent condition
  picture.naming$incongruent <<- do.trial(stimuli.list=c("cat","phouse"),
                                                  target.chunk="lhouse",SOA.list,nr.trials)
  if (plotting) lines(SOA.list,rowMeans(picture.naming$incongruent - neutral, na.rm=TRUE),
                      type="b", pch=15)


  ## word reading task, neutral condition
  word.reading$neutral <<- do.trial(stimuli.list="house",target.chunk="lhouse",
                                            SOA.list,nr.trials)
  if (plotting) {
    plot(SOA.list,SOA.list,type="n",ylim=c(-.1,.2), main="Word Reading", ylab="latency difference",
         xlab="SOA")
    abline(h=0, lty=2)
    neutral <- mean(word.reading$neutral, na.rm=TRUE)
  }
  
  ## concept congruent condition
  word.reading$concept.congruent <<- do.trial( stimuli.list=c("phouse","house"),
                                                     target.chunk="lhouse",SOA.list,nr.trials)
  if (plotting) lines(SOA.list,rowMeans(word.reading$concept.congruent, na.rm=TRUE) - neutral,
                      type="b", pch=17)
  
  ## category congruent condition
  word.reading$category.congruent <<- do.trial(stimuli.list=c("pcastle","house"),
                                                       target.chunk="lhouse",SOA.list,nr.trials)
  if (plotting) lines(SOA.list,rowMeans(word.reading$category.congruent, na.rm=TRUE) - neutral,
                      type="b", pch=16)

  ## incongruent condition
  word.reading$incongruent <<- do.trial(stimuli.list=c("pcat","house"),
                                               target.chunk="lhouse",SOA.list,nr.trials)
  if (plotting) lines(SOA.list,rowMeans(word.reading$incongruent, na.rm=TRUE) - neutral,
                      type="b", pch=15)

  ## write data to file
  if (!is.null(filename)) save(file=filename,actr,chunks,activation,evidence,
                               nr.trials,picture.naming,word.reading)
  
  ## print runtime of experiment
  print(cat("Run from ", date.and.time, " until ", date(), "\n"))
}

do.trial <- function(stimuli.list="house",target.chunk="lhouse",SOA.list=0,nr.trials=1, maxRT=2) {
  ## do one trial of memory retrieval
  ## stimuli.list takes either one or two chunks of the form c(distactor,target)

  RT <- matrix(NA,length(SOA.list),nr.trials,dimnames=list(SOA.list,NULL))
  SOA.list <- floor(SOA.list*actr$frequency)+1

  for (SOA in SOA.list) {
    duration <- abs(SOA)+maxRT*actr$frequency # default 2s (maxRT) after target stimulus onset a trial is ended
    init.model(duration, "DM.Rdat")
    min.onset <- 1
    number.chunks <- length(chunks$all)

    ## shift time axis and calculate stimuli onset
    if (length(stimuli.list)==2) {
      if (SOA<0) chunks$stimulus.onset <<- c(min.onset,abs(SOA))
      if (SOA>=0) chunks$stimulus.onset <<- c(SOA,min.onset)
    } else chunks$stimulus.onset <<- min.onset
    names(chunks$stimulus.onset) <<- stimuli.list

    for (trial in 1:nr.trials) {
      def.activation(duration, number.chunks)
      reset.chunks(duration,number.chunks)
      if (!(is.na(stimuli.list[2])) &&
          ## check if task is to name a picture with a facilitating distractor
          ## if so, flag is set to TRUE
          stimuli.list[1]==substr(stimuli.list[2],2,nchar(stimuli.list[2]))) {
        facilitator.flag <- FALSE
      } else facilitator.flag <- TRUE
      retrieve <- list(RT=0,chunk=NA) #tuple (timestep,chunk) returned by run.model

      ## while the target chunk is not yet retrieved:
      while ((!(target.chunk %in% retrieve$chunk) ||
              facilitator.flag==FALSE) &&
             retrieve$RT<duration) {
        ## if condition is facilitatory (flag=FALSE), and target chunk has crossed
        ## the context threshold before target stimulus onset, go on and set flag to TRUE
        if (facilitator.flag==FALSE &&
            target.chunk %in% retrieve$chunk &&
            retrieve$RT>max(chunks$stimulus.onset) &&
            SOA<0) {
          facilitator.flag <- TRUE
        }

        ## run the RACE equation until next context threshold crossing
        retrieve <- run.model(retrieve$RT+1,duration, target.chunk)

    }

      ## calculate latency for this trial and SOA
      RT[which(SOA.list==SOA),trial] <-
        latency(activation$all[target.chunk,],
                ret.onset=chunks$stimulus.onset[length(chunks$stimulus.onset)])
    }
  }  
  return(RT)
}

